# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.


class Asynchronous:
    def __init__(self, **kwargs):
        pass

    async def async_read_blob(self, *, blob_name, pool, statistics, **kwargs):
        pass
