"""
Draken Morsels Module

Morsel (batch) data structures for columnar processing.
"""

from opteryx.draken.morsels.align import align_tables
from opteryx.draken.morsels.align import align_tables_pyarray

__all__ = ["align_tables", "align_tables_pyarray"]
