# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Join operations module.

All individual .pyx files are compiled into a single .so file.
Import functions directly from opteryx.compiled.joins
"""

# This is created by the setup.py build process
from .join_definitions import *  # noqa: F403,F401
